# How many seconds to wait before re-fetching data from a unit
CACHE_INTERVAL_SECONDS = 5

# Magic related to generating the auth tokens
W_PARAM = bytearray.fromhex('44c73283b498d432ff25f5c8e06a016aef931e68f0a00ea710e36e6338fb22db')
S_PARAM = 0

# Default timeouts for interacting with the units
UNIT_CONNECT_TIMEOUT_SECONDS = 1.2
UNIT_RESPONSE_TIMEOUT_SECONDS = 8.0

# Default timeouts for interacting with the Kumo Cloud
KUMO_CONNECT_TIMEOUT_SECONDS = 5
KUMO_RESPONSE_TIMEOUT_SECONDS = 60

POSSIBLE_SENSORS = 4



""" Class used to represent the Kumo Cloud service
"""

import json
import logging
import re
import requests
from urllib3.util.retry import Retry
from requests.exceptions import Timeout
from getpass import getpass
from .py_kumo import PyKumo
from .py_kumo_base import PyKumoBase
from .py_kumo_station import PyKumoStation
from .const import KUMO_CONNECT_TIMEOUT_SECONDS, KUMO_RESPONSE_TIMEOUT_SECONDS

_LOGGER = logging.getLogger(__name__)

KUMO_UNIT_TYPE_TO_CLASS = {
    "ductless": PyKumo,
    "mvz": PyKumo,
    "sez": PyKumo,
    "pead": PyKumo,
    "headless": PyKumoStation
}

class KumoCloudAccount:
    """ API to talk to KumoCloud servers
    """
    def __init__(self, username, password, kumo_dict=None):
        """ Constructor from URL
        """
        if kumo_dict:
            self._url = None
            self._kumo_dict = kumo_dict
            self._need_fetch = False
            self._username = None
            self._password = None
            self._units = {}
        else:
            self._url = "https://geo-c.kumocloud.com/login"
            self._kumo_dict = None
            self._need_fetch = True
            self._username = username
            self._password = password
            self._units = {}

    @staticmethod
    def Factory(username=None, password=None):
        """Factory that prompts for username/password if not given
        """
        if username is None:
            username = input('Kumo Cloud username: ')
        if password is None:
            password = getpass()

        return KumoCloudAccount(username, password)

    @staticmethod
    def _parse_unit(raw_unit):
        """ Parse needed fields from raw json and return dict representing
            a unit
        """
        unit = {}
        fields = {'serial', 'label', 'address', 'password', 'cryptoSerial', 'mac', 'unitType'}
        try:
            for field in fields:
                unit[field] = raw_unit[field]
        except KeyError:
            pass
        return unit

    def _fetch_if_needed(self):
        """ Fetch configuration from server.
        """
        if self._url and self._need_fetch:
            headers = {'Accept': 'application/json, text/plain, */*',
                       'Accept-Encoding': 'gzip, deflate, br',
                       'Accept-Language': 'en-US,en',
                       'Content-Type': 'application/json'}

            body = { "username": self._username,
                     "password": self._password,
                     "appVersion": "2.2.0" }
            try:
                response = requests.post(self._url, headers=headers, data=json.dumps(body),
                                         timeout=(KUMO_CONNECT_TIMEOUT_SECONDS,
                                                  KUMO_RESPONSE_TIMEOUT_SECONDS))
            except Timeout as ex:
                response = None
                _LOGGER.warning("Timeout querying KumoCloud: %s", str(ex))
            if response:
                if response.ok:
                    self._kumo_dict = response.json()
                else:
                    _LOGGER.warning("Error response from KumoCloud: %s %s}",
                                    str(response.status_code), response.text)
            # Only try to fetch once
            self._need_fetch = False
            if not self._kumo_dict:
                _LOGGER.warning("No JSON returned from KumoCloud; check credentials?")
                return

        self._units = {}
        try:
            for child in self._kumo_dict[2]['children']:
                for raw_unit in child['zoneTable'].values():
                    unit = self._parse_unit(raw_unit)
                    serial = unit['serial']
                    self._units[serial] = unit
                if 'children' in child:
                    for grandchild in child['children']:
                        for raw_unit in grandchild['zoneTable'].values():
                            unit = self._parse_unit(raw_unit)
                            serial = unit['serial']
                            self._units[serial] = unit
        except KeyError:
            pass

    def try_setup(self):
        """Try to set up and return success/failure"""
        self._fetch_if_needed()

        return len(self._units.keys()) > 0

    def get_raw_json(self):
        """Return raw dict retrieved from KumoCloud"""
        return self._kumo_dict

    def get_all_units(self):
        """ Return list of all unit serial numbers
        """
        self._fetch_if_needed()

        return self._units.keys()

    def get_indoor_units(self):
        """ Return list of indoor unit serial numbers
        """

        return list(filter(
            lambda x: self.is_indoor_unit(x), self.get_all_units()))

    def get_kumo_stations(self):
        """ Return list of kumo station serial numbers
        """

        return list(filter(
            lambda x: self.is_headless_unit(x), self.get_all_units()))

    def get_name(self, unit):
        """ Return name of given unit
        """
        self._fetch_if_needed()

        try:
            return self._units[unit]['label']

        except KeyError:
            pass

        return None

    def is_indoor_unit(self, unit):
        """ Return whether unit is an indoor unit
        """
        self._fetch_if_needed()

        unit_type = self.get_unit_type(unit)
        if unit_type == "headless":
            return False
        return True

    def is_headless_unit(self, unit):
        """ Return whether unit is a headless unit (KumoStation)
        """
        self._fetch_if_needed()

        unit_type = self.get_unit_type(unit)
        if unit_type == "headless":
            return True
        return False

    def get_unit_type(self, unit):
        """ Return unit type of given unit
        """
        self._fetch_if_needed()

        try:
            return self._units[unit]['unitType']

        except KeyError:
            pass

        return None

    def get_address(self, unit):
        """ Return IP address of named unit
        """
        self._fetch_if_needed()

        try:
            return self._units[unit]['address']

        except KeyError:
            pass

        return None

    def get_mac(self, unit):
        """ Return mac address of named unit
        """
        self._fetch_if_needed()

        try:
            return self._units[unit]['mac']

        except KeyError:
            pass

        return None

    def get_credentials(self, unit):
        """ Return dict of credentials required to talk to unit
        """
        self._fetch_if_needed()

        try:
            credentials = {'password': self._units[unit]['password'],
                           'crypto_serial': self._units[unit]['cryptoSerial']}
            return credentials

        except KeyError:
            pass

        return None

    def make_pykumos(self, timeouts=None, init_update_status=True):
        """ Return a dict mapping names of all indoor units to newly-created
        `PyKumoBase` objects
        """
        kumos = {}
        for unitSerial in list(self.get_all_units()):
            name = self.get_name(unitSerial)
            if name in kumos:
                # I'm not sure if it's possible to have the same name repeated,
                # but just in case...
                m = re.match(r'(.*) \(([0-9]*)\)', name)
                if m:
                    name = m.group(1) + ' ({})'.format(int(m.group(2)) + 1)
                else:
                    kumos[name + ' (1)'] = kumos.pop(name)
                    name = name + ' (2)'
                # results in a name like "A/C unit (2)"
            unitType = self.get_unit_type(unitSerial)

            kumo_class = KUMO_UNIT_TYPE_TO_CLASS.get(unitType, PyKumo)
            kumos[name] = kumo_class(name, self.get_address(unitSerial),
                    self.get_credentials(unitSerial), timeouts, unitSerial)

        if init_update_status:
            for pk in kumos.values():
                pk.update_status()

        return kumos



""" Class used to represent indoor units
"""

import hashlib
import base64
import time
import logging
import requests
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from requests.exceptions import Timeout
from getpass import getpass
from .const import CACHE_INTERVAL_SECONDS, W_PARAM, S_PARAM, UNIT_CONNECT_TIMEOUT_SECONDS, UNIT_RESPONSE_TIMEOUT_SECONDS

_LOGGER = logging.getLogger(__name__)

class PyKumoBase:
    """ Talk to and control one indoor unit.
    """
    # pylint: disable=R0904, R0902

    def __init__(self, name, addr, cfg_json, timeouts=None, serial=None):
        """ Constructor
        """
        self._name = name
        self._address = addr
        self._serial = serial
        self._security = {
            'password': base64.b64decode(cfg_json["password"]),
            'crypto_serial': bytearray.fromhex(cfg_json["crypto_serial"])}
        if not timeouts:
            _LOGGER.info("Use default timeouts")
            self._timeouts = (UNIT_CONNECT_TIMEOUT_SECONDS,
                              UNIT_RESPONSE_TIMEOUT_SECONDS)
        else:
            _LOGGER.info("Use timeouts=%s", str(timeouts))
            connect_timeout = timeouts[0] if timeouts[0] else UNIT_CONNECT_TIMEOUT_SECONDS
            response_timeout = timeouts[1] if timeouts[1] else UNIT_RESPONSE_TIMEOUT_SECONDS
            self._timeouts = (connect_timeout, response_timeout)
        self._status = {}
        self._profile = {}
        self._sensors = []
        self._last_status_update = time.monotonic() - 2 * CACHE_INTERVAL_SECONDS

    def _token(self, post_data):
        """ Compute URL including security token for a given command
        """
        data_hash = hashlib.sha256(self._security['password'] +
                                   post_data).digest()

        intermediate = bytearray(88)
        intermediate[0:32] = W_PARAM[0:32]
        intermediate[32:64] = data_hash[0:32]
        intermediate[64:66] = bytearray.fromhex("0840")
        intermediate[66] = S_PARAM
        intermediate[79] = self._security['crypto_serial'][8]
        intermediate[80:84] = self._security['crypto_serial'][4:8]
        intermediate[84:88] = self._security['crypto_serial'][0:4]

        token = hashlib.sha256(intermediate).hexdigest()

        return token

    def _request(self, post_data):
        """ Send request to configured unit and return response dict
        """
        url = "http://" + self._address + "/api"
        token = self._token(post_data)
        headers = {'Accept': 'application/json, text/plain, */*',
                   'Content-Type': 'application/json'}
        token_param = {'m': token}
        try:
            with requests.Session() as session:
                retries = Retry(total=3)
                session.mount('http://', HTTPAdapter(max_retries=retries))
                _LOGGER.debug("Issue request %s %s", url, post_data)
                response = session.put(
                    url, headers=headers, data=post_data, params=token_param,
                    timeout=self._timeouts)
                return response.json()            
        except Timeout as ex:
            _LOGGER.warning("Timeout issuing request %s: %s", url, str(ex))
        except Exception as ex:
            _LOGGER.warning("Error issuing request %s: %s", url, str(ex))
        return {}

    def get_status(self):
        """ Last retrieved status dictionary from unit """
        return self._status

    def update_status(self):
        """ Retrieve and cache current status dictionary if enough time
            has passed
        """
        raise NotImplementedError()

    def get_name(self):
        """ Unit's name """
        return self._name

    def get_serial(self):
        """ Unit's serial number """
        return self._serial

    def get_sensor_rssi(self):
        """ Last retrievd sensor signal strength, if any """
        val = None
        try:
            for sensor in self._sensors:
                if sensor['rssi'] is not None:
                    return sensor['rssi']
        except KeyError:
            val = None
        return val

    def get_wifi_rssi(self):
        """ Last retrieved WiFi signal strengh, if any """
        val = None
        try:
            val = self._profile['wifiRSSI']
        except KeyError:
            val = None
        return val
